//! This file is part of Flight.
//! 
//! Flight is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
//! 
//! Flight is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
//! You should have received a copy of the GNU General Public License along with Flight. If not, see <https://www.gnu.org/licenses/>.

use super::ffi_gles::*;
use super::matrix3::Matrix3;
use super::rotation::Rotation;
use super::shader::{ActiveShader, Shader, Uniform};
use super::vector2::{F2, f2};
use super::vector3::{F3, f3};
use super::vertex::VertexBinding;
use super::window::Window;

/// A point of view that can be placed in the world.
/// This type keeps track of:
/// - viewpoint location and orientation
/// - parameters for perspective projection
/// - some cached transformations (for efficiency)
///
/// In normal rendering, these transformations happen:
/// (model coordinates)
/// - model scale
/// - model rotate
/// - model translate
/// (world coordinates)
/// - camera translate
/// - camera rotate
/// (view coordinates)
/// - x and y clip transform
/// - z clip transform
/// - perspective w divide (in hardware)
/// (clip coordinates)
/// - viewport scaling (in hardware)
/// (screen coordinates)
/// See: http://www.songho.ca/opengl/gl_projectionmatrix.html
///
/// The particular approach chosen for this code is designed for performance.
/// Whether it actually makes sense... yeah I don't really know.
///
/// The camera and projection help take care of calculating all of these
/// transformation steps.
pub struct Camera {
	/// See `CameraUniforms::clip_matrix`.
	clip_matrix: Matrix3,
	/// See `CameraUniforms::clip_translate`.
	clip_translate: F3,
	/// Orientation of camera.
	rotate: Rotation,
	/// Position of camera.
	translate: F3,
	/// Scale to convert from view coordinates to clip coordinates (0..1).
	clip_scale: F2,
	/// Rectangle in which view is drawn in window coordinates (pixels).
	/// (0, 0) is the bottom left of the window.
	/// See also `viewport_topleft`.
	viewport_size: F2,
	/// See `viewport_size`.
	viewport_topleft: F2,
	/// See `CameraUniforms::clip_z_offset`.
	z_offset: f32,
	/// See `CameraUniforms::clip_z_scale`.
	z_scale: f32,
}
impl Camera {
	/// `near` is the minimum distance objects have to be in front of the camera
	/// to be visible.
	/// `far` is the maximum distance objects can be in front of the camera to be
	/// visible.
	/// `fov` is the vertical field of view in radians.
	/// The horizontal FoV is scaled by the aspect ratio of the viewport; wide
	/// viewport shapes will have extra-wide FoV horizontally.
	/// `viewport_topleft` is the location of the view rectangle in pixels.
	/// `viewport_size` is the size of the view rectangle in pixels.
	pub fn new(near: f32, far: f32, fov: f32, viewport_topleft: F2, viewport_size: F2) -> Self {
		let half_frustum_base = (fov*0.5).tan()*near;
		let mut out = Self {
			clip_matrix: Matrix3::identity(),
			clip_translate: f3(0.0, 0.0, 0.0),
			rotate: Rotation::identity(),
			translate: f3(0.0, 0.0, 0.0),
			clip_scale: f2(0.0, near/half_frustum_base),
			viewport_size,
			viewport_topleft,
			z_scale: (far+near)/(far-near),
			z_offset: -2.0*near*far/(far-near),
		};
		out.viewport(viewport_topleft, viewport_size);
		out
	}
	/// Change view rectangle and update projection to match new aspect ratio.
	pub fn viewport(&mut self, topleft: F2, size: F2) {
		self.viewport_topleft = topleft;
		self.viewport_size = size;
		let aspect_ratio = if size.y <= 0.0 { 1.0 } else { size.x/size.y };
		self.clip_scale.x = self.clip_scale.y/aspect_ratio;
		// Recalculate cache for new aspect ratio.
		self.place(self.translate, self.rotate);
	}
	/// Change position and orientation of camera in the world.
	/// `translate` - location of centre of camera in world coordinates.
	/// `rotate` - orientation of camera relative to pointing along world z+.
	pub fn place(&mut self, translate: F3, rotate: Rotation) {
		self.rotate = rotate;
		self.translate = translate;
		// Cache world to clip transform for drawing `Anchor`s.
		let view_matrix = rotate.inverse().to_matrix3();
		let view_translate = view_matrix.apply(-translate);
		let (clip_matrix, clip_translate) = CameraUniforms::perspective(view_matrix, view_translate, self.clip_scale);
		self.clip_matrix = clip_matrix;
		self.clip_translate = clip_translate;
	}
	/// Set current viewport to camera's view rectangle.
	pub fn draw(&self, window: &mut Window) { unsafe {
		let gl = window.gl();
		(gl.Viewport)(
			self.viewport_topleft.x as GLint,
			self.viewport_topleft.y as GLint,
			self.viewport_size.x as GLsizei,
			self.viewport_size.y as GLsizei,
		);
	} }
	pub fn translate(&self) -> F3 { self.translate }
	pub fn rotate(&self) -> Rotation { self.rotate }
}

/// Stored locations of the uniforms a shader uses to accept a projection.
pub struct CameraUniforms {
	/// Matrix that applies model scale, model rotate, camera rotate, and part of
	/// x and y clip transform.
	/// Applied before `clip_translate`.
	pub clip_matrix: Option<Uniform>,
	/// Translation that applies model translate, camera translate, and part of x
	/// and y clip transform.
	/// Applied after `clip_matrix`.
	pub clip_translate: Option<Uniform>,
	/// Values to multiply then add with z to convert from view coordinates to
	/// clip coordinates (0..1) when combined with `clip_z_scale`.
	pub clip_z_offset: Option<Uniform>,
	/// See `clip_z_offset`.
	pub clip_z_scale: Option<Uniform>,
	/// Matrix that applies model scale and model rotate.
	/// This is useful for converting normals to world coordinates to calculate
	/// lighting.
	pub light_matrix: Option<Uniform>,
}
impl CameraUniforms {
	/// Get uniforms from a shader.
	/// The `light_matrix` uniform is only included if `with_light` is `true`.
	/// (This can be used to avoid debug warnings for shaders that don't use it)
	pub fn new<Vertex: VertexBinding>(window: &mut Window, shader: &Shader<Vertex>, with_light: bool) -> Self {
		Self {
			clip_matrix: shader.uniform(window, "clip_matrix"),
			clip_translate: shader.uniform(window, "clip_translate"),
			clip_z_offset: shader.uniform(window, "clip_z_offset"),
			clip_z_scale: shader.uniform(window, "clip_z_scale"),
			light_matrix: if with_light { shader.uniform(window, "light_matrix") } else { None },
		}
	}
	/// Set the projection for an active shader by calculating it from camera and
	/// model transforms.
	/// The `light_matrix` uniform is only calculated and set if `with_light` is `true`.
	/// (This saves compute overhead when it isn't needed)
	///
	/// This function is a straightforward way to setup the projection, but may
	/// wasteful in cases like:
	/// - the model transform doesn't change more than once every >= 3 frames
	/// - the model transform needs to be applied in >= 2 other operations every
	///   frame
	/// For these cases, see `Anchor`.
	pub fn set<Vertex: VertexBinding>(&self, window: &mut Window, shader: &ActiveShader<Vertex>,
		camera: &Camera, model_scale: F3, model_rotate: Rotation, model_translate: F3, with_light: bool,
	) {
		// Combine both rotations while they're quarternions and we can do it
		// efficiently.
		let camera_inverse = camera.rotate.inverse();
		let rotate = model_rotate.then(camera_inverse);

		// Convert model scale and both rotations into a 3x3 matrix.
		let view_matrix = rotate.to_matrix3_scale(model_scale);

		// Combine translations and apply camera rotation so they can just be scaled
		// and added to pre-clip coordinates.
		let view_translate = camera_inverse.apply(model_translate-camera.translate);

		let (clip_matrix, clip_translate) = Self::perspective(
			view_matrix, view_translate,
			camera.clip_scale);

		shader.set(window, &self.clip_matrix, &clip_matrix);
		shader.set(window, &self.clip_translate, clip_translate);
		shader.set(window, &self.clip_z_offset, camera.z_offset);
		shader.set(window, &self.clip_z_scale, camera.z_scale);

		if with_light {
			let light_matrix = model_rotate.to_matrix3_scale(model_scale);
			shader.set(window, &self.light_matrix, &light_matrix);
		}
	}
	/// Set the projection for an active shader by calculating it from `Camera`
	/// and `Anchor` transforms.
	/// The `light_matrix` uniform is only calculated and set if `with_light` is `true`.
	/// (This doesn't save any compute overhead in this case)
	pub fn set_anchor<Vertex: VertexBinding>(&self, window: &mut Window, shader: &ActiveShader<Vertex>,
		camera: &Camera, anchor: &Anchor, with_light: bool,
	) {
		// Convert translation to clip coordinates so it can just be added.
		let clip_translate = camera.clip_matrix.apply(anchor.translate)+camera.clip_translate;

		// Combine model to world and world to clip transforms.
		let clip_matrix = anchor.scale_rotate.then(&camera.clip_matrix);

		shader.set(window, &self.clip_matrix, &clip_matrix);
		shader.set(window, &self.clip_translate, clip_translate);
		shader.set(window, &self.clip_z_offset, camera.z_offset);
		shader.set(window, &self.clip_z_scale, camera.z_scale);

		if with_light {
			shader.set(window, &self.light_matrix, &anchor.scale_rotate);
		}
	}
	/// Set the projection for an active shader by calculating it just from the
	/// `Camera` transform.
	/// This is useful for saving compute when drawing meshes with no transform of
	/// their own, an optimisation particularly useful for `SceneryMesh`.
	pub fn set_identity<Vertex: VertexBinding>(&self, window: &mut Window, shader: &ActiveShader<Vertex>, camera: &Camera, with_light: bool) {
		shader.set(window, &self.clip_matrix, &camera.clip_matrix);
		shader.set(window, &self.clip_translate, camera.clip_translate);
		shader.set(window, &self.clip_z_offset, camera.z_offset);
		shader.set(window, &self.clip_z_scale, camera.z_scale);

		if with_light {
			shader.set(window, &self.light_matrix, &Matrix3::identity());
		}
	}

	/// Build a perspective projection from an input to view transform and some
	/// cached parameters.
	fn perspective(view_matrix: Matrix3, view_translate: F3, clip_scale: F2) -> (Matrix3, F3) {
		let mut clip_matrix = view_matrix;
		// Add scaling of x any y to convert them to clip space.
		// We need the original view z after applying the matrix, so we convert z
		// separately.
		//
		// The overall transformation we want to do here is project the x, y
		// position along the line to the camera onto the near plane (small face of
		// of the camera frustum), then squash them into clip range (-1 to 1).
		// The near plane of the frustum becomes the front of the clip cube, the far
		// plane becomes the equally-sized back face.
		// The GPU automatically divides x, y, and z by w, and we need to set w to
		// the original view z to get perspective, so we'll need to allow for that.
		// We end up with these formulae:
		// x_projected = x_view*near/z_view (project onto near plane)
		// x_clip = x_projected/half_frustum_base
		// x_preclip = x_clip*z_view (compensate for the automatic divide)
		//           = x_projected/half_frustum_base*z_view
		//           = x_view*near/z_view/half_frustum_base*z_view
		//           = x_view*near/half_frustum_base
		// x_scale = near/half_frustum_base
		// x_preclip = x_view*x_scale
		//
		// `x_scale` and `y_scale` become `clip_scale`, cached in `Camera`.
		clip_matrix.matrix[0] *= clip_scale.x;
		clip_matrix.matrix[1] *= clip_scale.y;
		// No op: clip_matrix.matrix[2] *= 1.0;
		clip_matrix.matrix[3] *= clip_scale.x;
		clip_matrix.matrix[4] *= clip_scale.y;
		// No op: clip_matrix.matrix[5] *= 1.0;
		clip_matrix.matrix[6] *= clip_scale.x;
		clip_matrix.matrix[7] *= clip_scale.y;
		// No op: clip_matrix.matrix[8] *= 1.0;

		// Calculate conversion of z to clip space to go after the view matrix.
		// In the vertex shader we'll multiply and then add because that neatly
		// ends up as one madd instruction (I believe?).
		// Because of the automatic divide by z_view, and because non-linear depth
		// makes better use of available precision, we want to end up with something
		// this shape:
		// z_clip = (z_view*z_scale+z_offset)/z_view
		// z_preclip = z_clip*z_view = z_view*z_scale+z_offset
		// We know the bounds are these:
		// z_clip = -1 when z_view = near
		// z_clip = 1 when z_view = far
		// So:
		// -1 = (near*z_scale+z_offset)/near ... 1 = (far*z_scale+z_offset)/far
		// -near = near*z_scale+z_offset     ... far = far*z_scale+z_offset
		// z_offset = -near*z_scale-near     ... z_offset = far-far*z_scale
		// -near*z_scale-near = far-far*z_scale
		// z_scale*(far-near) = far+near
		// z_scale = (far+near)/(far-near)
		// z_offset = -near*(far+near)/(far-near)-near
		//          = -near*((far-near)+(far+near))/(far-near)
		//          = -2*near*far/(far-near)
		// These are cached in `Camera`.

		// Apply clip transform to translation so it can be added after the matrix
		// multiplication.
		// This saves some redundant arithmetic compared to using the traditional
		// 4x4 matrix and 4 element vector.
		// This is also faster then building separate matrices, using one of them to
		// adjust the translation, then multiplying them both together (most likely)
		// for constantly moving objects without a light matrix.
		let clip_translate = f3(
			view_translate.x*clip_scale.x,
			view_translate.y*clip_scale.y,
			view_translate.z,
		);

		(clip_matrix, clip_translate)
	}
}

/// Tracks the transform of an object in the world and caches transformation
/// parameters.
/// This can be used to get better performance when drawing objects whose
/// transform doesn't change more than once every >=3 frames.
#[derive(Clone)]
pub struct Anchor {
	/// Cached model scale and model rotate.
	pub scale_rotate: Matrix3,
	/// Cached model translate in world coordinates.
	/// Since this is stored unchanged, it can be directly modified.
	pub translate: F3,
}
impl Anchor {
	pub fn new(scale: F3, rotate: Rotation, translate: F3) -> Self {
		Self {
			scale_rotate: rotate.to_matrix3_scale(scale),
			translate,
		}
	}
}
