//! This file is part of Flight.
//! 
//! Flight is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
//! 
//! Flight is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
//! You should have received a copy of the GNU General Public License along with Flight. If not, see <https://www.gnu.org/licenses/>.

use std::convert::TryFrom;
use std::marker::PhantomData;
use std::rc::Weak;
use super::ffi_gles::*;
use super::metrics::MetricUse;
use super::shader::ActiveShader;
use super::vertex::{INDEX_TYPE, Index, Primitive, VertexBinding};
use super::window::Window;

/// A mesh who's geometry does not change once it has been created.
/// It can be moved around, rotated, and scaled, but that's it.
pub struct StaticMesh<Vertex: VertexBinding> {
	/// Handle of index buffer.
	indices: GLuint,
	/// Bytes used on GPU.
	#[allow(dead_code)]
	geometry_bytes: MetricUse,
	gl: Weak<GL>,
	/// Number of indices in index buffer.
	length: usize,
	phantom: PhantomData<Vertex>,
	/// What to draw.
	primitive: Primitive,
	/// Handle of vertex buffer.
	vertices: GLuint,
}
impl<Vertex: VertexBinding> StaticMesh<Vertex> {
	/// Load a mesh onto the GPU.
	pub fn new(window: &mut Window, primitive: Primitive, indices: &[Index], vertices: &[Vertex]) -> Self {
		let geometry_bytes = window.metrics().geometry_bytes.take(
			std::mem::size_of::<Index>()*indices.len()+std::mem::size_of::<Vertex>()*vertices.len());


		debug_assert!(indices.len()%primitive.vertices() == 0);
		let length = indices.len();
		let (indices, vertices) = unsafe {
			let gl = window.gl();
			let vertices = Self::upload_buffer(gl, GL_ARRAY_BUFFER, &vertices);
			let indices = Self::upload_buffer(gl, GL_ELEMENT_ARRAY_BUFFER, &indices);
			(indices, vertices)
		};

		Self {
			indices,
			geometry_bytes,
			gl: window.weak_gl(),
			length,
			phantom: PhantomData,
			primitive,
			vertices,
		}
	}
	/// Draw mesh using currently active shader.
	pub fn draw(&self, window: &mut Window, _shader: &ActiveShader<Vertex>) { unsafe {
		let gl = window.get_gl(&self.gl);

		// Activate buffers.
		(gl.BindBuffer)(GL_ARRAY_BUFFER, self.vertices);
		(gl.BindBuffer)(GL_ELEMENT_ARRAY_BUFFER, self.indices);
		Vertex::load(gl);

		// Draw mesh.
		(gl.DrawElements)(self.primitive.raw(), self.length as GLsizei, INDEX_TYPE, std::ptr::null());

		// Tidy up.
		if cfg!(debug_assertions) {
			(gl.BindBuffer)(GL_ARRAY_BUFFER, 0);
			(gl.BindBuffer)(GL_ELEMENT_ARRAY_BUFFER, 0);
		}
		gl.check();
	} }
	/// Create and populate a static vertex or index buffer.
	/// Mostly intended for internal use.
	pub unsafe fn upload_buffer<T>(gl: &GL, mode: GLenum, data: &[T]) -> GLuint {
		// Create and bind buffer.
		assert!(mode == GL_ARRAY_BUFFER || mode == GL_ELEMENT_ARRAY_BUFFER);
		let mut vbo = 0 as GLuint;
		(gl.GenBuffers)(1, &mut vbo);
		if vbo == 0 { panic!("GLES: failed to allocate VBO"); }
		(gl.BindBuffer)(mode, vbo);

		// Upload data.
		let size = (std::mem::size_of::<T>()*data.len()) as GLsizeiptr;
		(gl.BufferData)(mode, size, data.as_ptr() as *const GLvoid, GL_STATIC_DRAW);

		// Clear binding.
		(gl.BindBuffer)(mode, 0);
		gl.check();

		vbo
	}
}
impl<Vertex: VertexBinding> Drop for StaticMesh<Vertex> {
	fn drop(&mut self) { unsafe {
		if let Some(gl) = self.gl.upgrade() {
			let buffers = [self.indices, self.vertices];
			(gl.DeleteBuffers)(buffers.len() as GLsizei, buffers.as_ptr());
		}
	} }
}

/// A mesh that is generated on the fly every frame.
pub struct DynamicMesh<Vertex: VertexBinding> {
	/// Handle of index buffer.
	indices: GLuint,
	indices_data: Vec<Index>,
	/// Bytes used on GPU.
	geometry_bytes: MetricUse,
	gl: Weak<GL>,
	/// What to draw.
	primitive: Primitive,
	/// Handle of vertex buffer.
	vertices: GLuint,
	vertices_data: Vec<Vertex>,
}
impl<Vertex: VertexBinding> DynamicMesh<Vertex> {
	/// Create an empty mesh stream.
	pub fn new(window: &mut Window, primitive: Primitive) -> Self { unsafe {
		let geometry_bytes = window.metrics().geometry_bytes.take(0);

		let gl = window.gl();
		let mut buffers = [0 as GLuint; 2];
		(gl.GenBuffers)(buffers.len() as GLsizei, buffers.as_mut_ptr());
		if buffers.iter().any(|buffer| *buffer == 0) { panic!("GLES: failed to allocate VBO"); }

		Self {
			indices: buffers[0],
			indices_data: Vec::new(),
			geometry_bytes,
			gl: window.weak_gl(),
			primitive,
			vertices: buffers[1],
			vertices_data: Vec::new(),
		}
	} }
	/// Clear geometry (usually only useful with `draw_keep()`).
	pub fn clear(&mut self) {
		self.indices_data.clear();
		self.vertices_data.clear();
	}
	/// Add a vertex to the mesh and return it's index if there was space.
	pub fn push(&mut self, vertex: Vertex) -> Option<Index> {
		if let Ok(index) = Index::try_from(self.vertices_data.len()) {
			self.indices_data.push(index);
			self.vertices_data.push(vertex);
			Some(index)
		} else {
			None
		}
	}
	/// Repeat a previously added vertex.
	pub fn repeat(&mut self, index: Option<Index>) {
		if let Some(index) = index {
			self.indices_data.push(index);
		}
	}
	/// Draw mesh and clear it ready to add new geometry.
	pub fn draw(&mut self, window: &mut Window, shader: &ActiveShader<Vertex>) {
		self.draw_keep(window, shader);
		self.clear();
	}
	/// Draw mesh without clearing it, allowing it to be reused.
	/// This type is optimised for replacing the contents of the mesh every frame,
	/// so this mode may be sub-optimal.
	pub fn draw_keep(&mut self, window: &mut Window, _shader: &ActiveShader<Vertex>) { unsafe {
		self.geometry_bytes.update(
			std::mem::size_of::<Index>()*self.indices_data.len()+
			std::mem::size_of::<Vertex>()*self.vertices_data.len());

		let gl = window.get_gl(&self.gl);

		// Bind and load buffers.
		(gl.BindBuffer)(GL_ARRAY_BUFFER, self.vertices);
		let size = (std::mem::size_of::<Vertex>()*self.vertices_data.len()) as GLsizeiptr;
		(gl.BufferData)(GL_ARRAY_BUFFER, size, self.vertices_data.as_ptr() as *const GLvoid, GL_STREAM_DRAW);
		(gl.BindBuffer)(GL_ELEMENT_ARRAY_BUFFER, self.indices);
		let size = (std::mem::size_of::<Index>()*self.indices_data.len()) as GLsizeiptr;
		(gl.BufferData)(GL_ELEMENT_ARRAY_BUFFER, size, self.indices_data.as_ptr() as *const GLvoid, GL_STREAM_DRAW);
		Vertex::load(gl);

		// Draw mesh.
		let length = self.indices_data.len() as GLsizei;
		(gl.DrawElements)(self.primitive.raw(), length, INDEX_TYPE, std::ptr::null());

		// Tidy up.
		if cfg!(debug_assertions) {
			(gl.BindBuffer)(GL_ARRAY_BUFFER, 0);
			(gl.BindBuffer)(GL_ELEMENT_ARRAY_BUFFER, 0);
		}
		gl.check();
	} }
}
impl<Vertex: VertexBinding> Drop for DynamicMesh<Vertex> {
	fn drop(&mut self) { unsafe {
		if let Some(gl) = self.gl.upgrade() {
			let buffers = [self.indices, self.vertices];
			(gl.DeleteBuffers)(buffers.len() as GLsizei, buffers.as_ptr());
		}
	} }
}
