//! This file is part of Flight.
//! 
//! Flight is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
//! 
//! Flight is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
//! You should have received a copy of the GNU General Public License along with Flight. If not, see <https://www.gnu.org/licenses/>.

use std::cell::Cell;
use std::rc::{Rc, Weak};
use super::pretty_print;

/// Tracks amounts of resources used over a program lifecycle.
/// See `Window::metrics()` for access to this type for normal use.
pub struct Metrics {
	pub geometry_bytes: Metric,
	pub texture_bytes: Metric,
}
impl Metrics {
	pub fn new() -> Self {
		Self {
			geometry_bytes: Metric::new(),
			texture_bytes: Metric::new(),
		}
	}
	/// Print statistics at end of lifecycle.
	#[cfg(debug_assertions)]
	pub fn dump(&self) {
		let values = [
			("geometry (bytes)", format!("{}", pretty_print::Thousands(self.geometry_bytes.peak as u64))),
			("textures (bytes)", format!("{}", pretty_print::Thousands(self.texture_bytes.peak as u64))),
		];
		let label_bytes = values.iter().map(|(label, _)| label.as_bytes().len()).max().unwrap_or(0);
		let value_bytes = values.iter().map(|(_, value)| value.as_bytes().len()).max().unwrap_or(0);
		println!("Metrics (peak resource use):");
		for (label, value) in values.iter() {
			println!("- {:<label_pad$}: {:>value_pad$}",
				label, value, label_pad=label_bytes, value_pad=value_bytes);
		}
	}
	#[cfg(not(debug_assertions))]
	pub fn dump(&self) {}
}

pub struct Metric {
	#[cfg(debug_assertions)]
	value: Rc<Cell<usize>>,
	#[cfg(debug_assertions)]
	peak: usize,
}
impl Metric {
	fn new() -> Self {
		Self {
			#[cfg(debug_assertions)]
			value: Rc::new(Cell::new(0)),
			#[cfg(debug_assertions)]
			peak: 0,
		}
	}
	#[cfg(debug_assertions)]
	pub fn take(&mut self, amount: usize) -> MetricUse {
		let value = self.value.get()+amount;
		self.value.set(value);
		if value > self.peak { self.peak = value; }
		MetricUse {
			amount,
			value: Rc::downgrade(&self.value),
		}
	}
	#[cfg(not(debug_assertions))]
	pub fn take(&mut self, _amount: usize) -> MetricUse { MetricUse{} }
}

pub struct MetricUse {
	#[cfg(debug_assertions)]
	amount: usize,
	#[cfg(debug_assertions)]
	value: Weak<Cell<usize>>,
}
impl MetricUse {
	#[cfg(debug_assertions)]
	pub fn update(&mut self, amount: usize) {
		if let Some(value) = self.value.upgrade() {
			debug_assert!(value.get() >= self.amount);
			value.set(value.get()-self.amount+amount);
			self.amount = amount;
		}
	}
	#[cfg(not(debug_assertions))]
	pub fn update(&mut self, _amount: usize) {}
}
#[cfg(debug_assertions)]
impl Drop for MetricUse {
	fn drop(&mut self) {
		if let Some(value) = self.value.upgrade() {
			debug_assert!(value.get() >= self.amount);
			value.set(value.get()-self.amount);
		}
	}
}
