//! This file is part of Flight.
//! 
//! Flight is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
//! 
//! Flight is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
//! You should have received a copy of the GNU General Public License along with Flight. If not, see <https://www.gnu.org/licenses/>.

use std::ops::{Add, AddAssign, Div, DivAssign, Sub, SubAssign, Neg, Mul, MulAssign};

pub const fn f2(x: f32, y: f32) -> F2 {
	F2{x, y}
}

#[derive(Clone, Copy, Debug, PartialEq)]
pub struct F2 {
	pub x: f32,
	pub y: f32,
}
impl F2 {
	pub fn to_array(self) -> [f32; 2] {
		[self.x, self.y]
	}
	pub fn from_array(array: [f32; 2]) -> Self {
		Self{x: array[0], y: array[1]}
	}
	pub fn dot(self, other: Self) -> f32 {
		self.x*other.x+self.y*other.y
	}
	pub fn cross(self, other: Self) -> f32 {
		self.x*other.y-self.y*other.x
	}
	pub fn length(self) -> f32 {
		(self.x*self.x+self.y*self.y).sqrt()
	}
	pub fn length2(self) -> f32 {
		self.x*self.x+self.y*self.y
	}
	pub fn lerp(self, other: Self, factor: f32) -> Self {
		self+(other-self)*factor
	}
	pub fn unit(self) -> Option<Self> {
		let length2 = self.x*self.x+self.y*self.y;
		if length2 <= 0.0 { None } else { Some(self/length2.sqrt()) }
	}
	pub fn unit_zero(self) -> Self {
		self.unit().unwrap_or(Self{x: 0.0, y: 0.0})
	}

	pub fn abs(self) -> Self {
		Self{x: self.x.abs(), y: self.y.abs()}
	}
	pub fn ceil(self) -> Self {
		Self{x: self.x.ceil(), y: self.y.ceil()}
	}
	pub fn floor(self) -> Self {
		Self{x: self.x.floor(), y: self.y.floor()}
	}
	pub fn max(self, other: Self) -> Self {
		Self{x: self.x.max(other.x), y: self.y.max(other.y)}
	}
	pub fn min(self, other: Self) -> Self {
		Self{x: self.x.min(other.x), y: self.y.min(other.y)}
	}
	pub fn recip(self) -> Self {
		Self{x: self.x.recip(), y: self.y.recip()}
	}
	pub fn round(self) -> Self {
		Self{x: self.x.round(), y: self.y.round()}
	}
}
impl Add for F2 {
	type Output = Self;
	fn add(self, other: Self) -> Self {
		Self{x: self.x+other.x, y: self.y+other.y}
	}
}
impl AddAssign for F2 {
	fn add_assign(&mut self, other: Self) {
		self.x += other.x;
		self.y += other.y;
	}
}
impl std::fmt::Display for F2 {
	fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
		write!(f, "{},{}", self.x, self.y)
	}
}
impl Div<f32> for F2 {
	type Output = Self;
	fn div(self, other: f32) -> Self {
		self*other.recip()
	}
}
impl DivAssign<f32> for F2 {
	fn div_assign(&mut self, other: f32) {
		*self *= other.recip();
	}
}
impl Div<F2> for F2 {
	type Output = Self;
	fn div(self, other: Self) -> Self {
		self*other.recip()
	}
}
impl DivAssign<F2> for F2 {
	fn div_assign(&mut self, other: Self) {
		*self *= other.recip();
	}
}
impl Mul<f32> for F2 {
	type Output = Self;
	fn mul(self, other: f32) -> Self {
		Self{x: self.x*other, y: self.y*other}
	}
}
impl MulAssign<f32> for F2 {
	fn mul_assign(&mut self, other: f32) {
		self.x *= other;
		self.y *= other;
	}
}
impl Mul for F2 {
	type Output = Self;
	fn mul(self, other: Self) -> Self {
		Self{x: self.x*other.x, y: self.y*other.y}
	}
}
impl MulAssign for F2 {
	fn mul_assign(&mut self, other: Self) {
		self.x *= other.x;
		self.y *= other.y;
	}
}
impl Neg for F2 {
	type Output = Self;
	fn neg(self) -> Self {
		Self{x: -self.x, y: -self.y}
	}
}
impl Sub for F2 {
	type Output = Self;
	fn sub(self, other: Self) -> Self {
		Self{x: self.x-other.x, y: self.y-other.y}
	}
}
impl SubAssign for F2 {
	fn sub_assign(&mut self, other: Self) {
		self.x -= other.x;
		self.y -= other.y;
	}
}
