//! This file is part of Flight.
//! 
//! Flight is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
//! 
//! Flight is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
//! You should have received a copy of the GNU General Public License along with Flight. If not, see <https://www.gnu.org/licenses/>.

use collision::CollisionMesh;
use engine::{F3, f3, Rotation};

pub struct Surface {
	pub area: f32,
	pub centre: F3,
	pub normal: F3,
}

pub struct Aircraft {
	pub centre: F3,
	pub crashed_counter: u32,
	pub mass_recip: f32,
	pub rotation: Rotation,
	pub spin: F3,
	pub spin_damping: f32,
	pub spin_recip: f32,
	pub velocity: F3,
}
impl Aircraft {
	pub fn tick(&mut self, surfaces: &[Surface], floor: &CollisionMesh) {
		if self.crashed_counter > 0 {
			self.crashed_counter += 1;
			if self.crashed_counter > 60 {
				self.crashed_counter = 0;
				let direction = self.centre;
				let object = floor.ray(f3(0.0, 0.0, 0.0), direction, Some(1.0), None)
					.map(|hit| hit.point())
					.unwrap_or(self.centre);
				let radius = f3(0.0, self.centre.y, self.centre.z).unit().unwrap_or(f3(0.0, 1.0, 0.0));
				let face_in = self.rotation.apply(f3(0.0, 0.0, 1.0));
				let face_in = (face_in-radius*face_in.dot(radius)).unit().unwrap_or(f3(1.0, 0.0, 0.0));
				self.centre = object-direction.unit_zero()*10.0;
				self.rotation = Rotation::look(face_in, -radius);
				self.velocity = f3(0.0, 0.0, 0.0);
				self.spin = f3(0.0, 0.0, 0.0);
			}
			return;
		}
		assert!(self.velocity.x.is_finite() && self.velocity.y.is_finite() && self.velocity.z.is_finite());
		let length = self.velocity.length();
		if length > 1000.0 {
			self.velocity = self.velocity*(1000.0/length);
		}
		let mut forces = [(f3(0.0, 0.0, 0.0), f3(0.0, 0.0, 0.0)); 3];
		for (surface, force) in surfaces.iter().zip(forces.iter_mut()) {
			let normal = self.rotation.apply(surface.normal);
			assert!(normal.x.is_finite() && normal.y.is_finite() && normal.z.is_finite());
			let effect = self.velocity.dot(normal);
			assert!(effect.is_finite());
			if effect != 0.0 {
				let impulse = normal*(-effect*surface.area);
				let centre = self.rotation.apply(surface.centre);
				assert!(centre.x.is_finite() && centre.y.is_finite() && centre.z.is_finite());
				*force = (centre, impulse);
			}
		}
		for (centre, impulse) in forces.iter().copied() {
			self.linear_impulse(centre, impulse);
		}
		self.spin *= self.spin_damping;
		// See: https://www.gafferongames.com/post/physics_in_3d/
		let spin = Rotation{a: 0.0, b: self.spin.x*0.5, c: self.spin.y*0.5, d: self.spin.z*0.5};
		let delta = self.rotation.then(spin);
		self.rotation.a += delta.a;
		self.rotation.b += delta.b;
		self.rotation.c += delta.c;
		self.rotation.d += delta.d;
		self.rotation = self.rotation.unit_identity();
		self.centre += self.velocity*0.01;
		// TODO: use an anchor
		let matrix = self.rotation.to_matrix3();
		let rays = [
			(f3(0.0, 0.174, -0.702), f3(0.0, 0.094, 0.866)),
			(f3(-0.954, -0.02, -0.013), f3(0.954, -0.02, -0.013)),
		];
		for (start, end) in rays.iter().copied() {
			let start = matrix.apply(start)+self.centre;
			let end = matrix.apply(end)+self.centre;
			if floor.ray(start, end-start, Some(1.0), None).is_some() {
				self.crashed_counter = 1;
				break;
			}
		}
	}
	pub fn point_velocity(&self, world_offset: F3) -> F3 {
		// See: https://www.gafferongames.com/post/physics_in_3d/
		// See: https://github.com/godotengine/godot/blob/master/servers/physics_3d/godot_body_3d.h
		self.spin.cross(world_offset)+self.velocity
	}
	pub fn linear_impulse(&mut self, world_offset: F3, world_impulse: F3) {
		// See: https://github.com/godotengine/godot/blob/master/servers/physics_3d/godot_body_3d.h
		self.velocity += world_impulse*self.mass_recip;
		self.spin += world_offset.cross(world_impulse)*self.spin_recip;
	}
}
