//! This file is part of Flight.
//! 
//! Flight is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
//! 
//! Flight is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
//! You should have received a copy of the GNU General Public License along with Flight. If not, see <https://www.gnu.org/licenses/>.

use engine::{DynamicMesh, F3, f3, Index};
use std::convert::TryFrom;

pub trait CollisionVertex {
	fn position(&self) -> F3;
}
impl CollisionVertex for engine::Vertex {
	fn position(&self) -> F3 { F3::from_array(self.position) }
}

pub struct CollisionMesh {
	cell_size: F3,
	grid: Box<[Cell]>,
	grid_origin: F3,
	grid_size: [usize; 3],
}
impl CollisionMesh {
	pub fn invalid() -> Self {
		Self {
			cell_size: f3(0.0, 0.0, 0.0),
			grid: vec![].into_boxed_slice(),
			grid_origin: f3(0.0, 0.0, 0.0),
			grid_size: [0, 0, 0],
		}
	}
	pub fn new<Vertex: CollisionVertex>(indices: &[Index], vertices: &[Vertex], grid_size: [usize; 3]) -> Self {
		debug_assert!(grid_size[0] > 0);
		debug_assert!(grid_size[1] > 0);
		debug_assert!(grid_size[2] > 0);

		// We use `i32` cell coordinates later and we want to be able to step outside
		// the grid without overflow weirdness.
		debug_assert!(grid_size[0] <= usize::MAX-2 && i32::try_from(grid_size[0]+2).is_ok());
		debug_assert!(grid_size[1] <= usize::MAX-2 && i32::try_from(grid_size[1]+2).is_ok());
		debug_assert!(grid_size[2] <= usize::MAX-2 && i32::try_from(grid_size[2]+2).is_ok());

		debug_assert!(indices.len()%3 == 0);

		// Bounding box.
		let mut grid_min = vertices.first().map(|vertex| vertex.position()).unwrap_or(f3(0.0, 0.0, 0.0));
		let mut grid_max = grid_min;
		for vertex in vertices.iter() {
			let position = vertex.position();
			grid_min = position.min(grid_min);
			grid_max = position.max(grid_max);
		}

		// Group collision data by cell, using bounding boxes to approximate triangle
		// vs cell collision tests.
		let cell_size = f3(
			(grid_max.x-grid_min.x)/grid_size[0] as f32,
			(grid_max.y-grid_min.y)/grid_size[1] as f32,
			(grid_max.z-grid_min.z)/grid_size[2] as f32,
		);
		let length = grid_size[0]*grid_size[1]*grid_size[2];
		let mut triangles = vec![Vec::new(); length];
		for triangle in indices.chunks_exact(3) {
			let vertices = [
				vertices[triangle[0] as usize].position(),
				vertices[triangle[1] as usize].position(),
				vertices[triangle[2] as usize].position(),
			];
			let triangle_min = vertices[0].min(vertices[1]).min(vertices[2]);
			debug_assert!(triangle_min.x >= grid_min.x);
			debug_assert!(triangle_min.y >= grid_min.y);
			debug_assert!(triangle_min.z >= grid_min.z);
			let triangle_max = vertices[0].max(vertices[1]).max(vertices[2]);
			debug_assert!(triangle_max.x <= grid_max.x);
			debug_assert!(triangle_max.y <= grid_max.y);
			debug_assert!(triangle_max.z <= grid_max.z);
			let triangle = Triangle::new(vertices);
			let range_min = ((triangle_min-grid_min)/cell_size).floor();
			let range_min = [range_min.x as usize, range_min.y as usize, range_min.z as usize];
			let range_max = ((triangle_max-grid_min)/cell_size).floor();
			let range_max = [
				// Add one to make iteration easy, and clamp because triangle_max==grid_max
				// creates an edge case where a vertex may end up outside the grid.
				(range_max.x as usize+1).min(grid_size[0]),
				(range_max.y as usize+1).min(grid_size[1]),
				(range_max.z as usize+1).min(grid_size[2]),
			];
			for z in range_min[2]..range_max[2] {
				for y in range_min[1]..range_max[1] {
					for x in range_min[0]..range_max[0] {
						let index = x+y*grid_size[0]+z*grid_size[0]*grid_size[1];
						triangles[index].push(triangle.clone());
					}
				}
			}
		}

		let grid = triangles.into_iter()
			.map(|triangles: Vec<Triangle>| Cell{triangles: triangles.into_boxed_slice()})
			.collect::<Vec<Cell>>()
			.into_boxed_slice();

		Self {
			cell_size,
			grid,
			grid_origin: grid_min,
			grid_size,
		}
	}
	pub fn ray(&self, start: F3, direction: F3, range: Option<f32>, mut debug: Option<&mut DynamicMesh<engine::Vertex>>) -> Option<RayHit> {
		debug_assert!(self.grid_size != [0, 0, 0]);
		if direction.x+direction.y+direction.z == 0.0 { return None; }

		if cfg!(debug_assertions) {
			if let Some(debug) = debug.as_mut() {
				Self::debug_line(debug, start, start+direction*range.unwrap_or(100.0));
				let radius = f3(1.0, 1.0, 1.0);
				Self::debug_aabb(debug, start-radius, radius*2.0);
				Self::debug_aabb(debug, start+direction-radius, radius*2.0);
			}
		}

		// Move start of ray to just outside grid.
		// This should avoid overflow problems and walking useless cell positions
		// while not causing edge cases as much as using the exact edge of the grid.
		let aabb_origin = self.grid_origin-self.cell_size*0.5;
		let aabb_size = self.cell_size*f3(
			self.grid_size[0] as f32+1.0,
			self.grid_size[1] as f32+1.0,
			self.grid_size[2] as f32+1.0,
		);
		let start_time = Self::time_to_aabb(start, direction, aabb_origin, aabb_size);
		let start = start+direction*start_time;

		if cfg!(debug_assertions) {
			if let Some(debug) = debug.as_mut() {
				Self::debug_aabb(debug, aabb_origin, aabb_size);
				Self::debug_marker(debug, start);
			}
		}

		// Direction to move.
		let sign = [
			if direction.x >= 0.0 { 1 } else { -1 },
			if direction.y >= 0.0 { 1 } else { -1 },
			if direction.z >= 0.0 { 1 } else { -1 },
		];
		// Time it takes to step by one cell for each axis.
		let step = self.cell_size/direction.abs();
		// Index of current cell.
		let mut cell = [
			((start.x-self.grid_origin.x)/self.cell_size.x).floor() as i32,
			((start.y-self.grid_origin.y)/self.cell_size.y).floor() as i32,
			((start.z-self.grid_origin.z)/self.cell_size.z).floor() as i32,
		];
		// Time until stepping along each axis.
		let mut next = f3(
			Self::time_to_boundary(start.x, direction.x, self.grid_origin.x, self.cell_size.x),
			Self::time_to_boundary(start.y, direction.y, self.grid_origin.y, self.cell_size.y),
			Self::time_to_boundary(start.z, direction.z, self.grid_origin.z, self.cell_size.z),
		);
		// Time spent tracing the ray.
		let mut time = start_time;
		// Earliest collision found.
		// We have to keep this across multiple cells because a large triangle in a
		// cell we try earlier can indicate a collision, and if we immediately stop
		// we might miss a closer collision with a tiny triangle in a later cell.
		let mut best_hit = None;
		let mut best_time = -1.0;

		// Step through cells the ray touches.
		loop {
			// Stop if already beyond range.
			if range.is_some() && time > range.unwrap() {
				return best_hit;
			}
			// Check for collisions with triangles within current cell.
			let inside = cell[0] >= 0 && cell[0] < self.grid_size[0] as i32 &&
				cell[1] >= 0 && cell[1] < self.grid_size[1] as i32 &&
				cell[2] >= 0 && cell[2] < self.grid_size[2] as i32;
			if inside {
				if cfg!(debug_assertions) {
					if let Some(debug) = debug.as_mut() {
						Self::debug_aabb(debug, self.grid_origin+self.cell_size*f3(
							cell[0] as f32, cell[1] as f32, cell[2] as f32), self.cell_size);
					}
				}
				let index = cell[0] as usize +
					cell[1] as usize*self.grid_size[0] +
					cell[2] as usize*self.grid_size[0]*self.grid_size[1];
				debug_assert!(index < self.grid.len());
				for triangle in self.grid[index].triangles.iter() {
					#[cfg(debug_assertions)]
					if let Some(debug) = debug.as_mut() {
						Self::debug_triangle(debug, triangle.vertices);
					}
					if let Some(hit) = triangle.ray(start, direction, range) {
						debug_assert!(hit.time >= 0.0);
						debug_assert!(range.is_none() || hit.time <= range.unwrap());
						if cfg!(debug_assertions) {
							if let Some(debug) = debug.as_mut() {
								Self::debug_marker(debug, hit.point());
							}
						}
						if best_hit.is_none() || hit.time < best_time {
							best_time = hit.time;
							best_hit = Some(hit);
						}
					}
				}
			} else if (direction.x >= 0.0 && cell[0] >= self.grid_size[0] as i32) ||
				(direction.x < 0.0 && cell[0] < 0) ||
				(direction.y >= 0.0 && cell[1] >= self.grid_size[1] as i32) ||
				(direction.y < 0.0 && cell[1] < 0) ||
				(direction.z >= 0.0 && cell[2] >= self.grid_size[2] as i32) ||
				(direction.z < 0.0 && cell[2] < 0)
			{
				// Left the mesh, finish.
				return best_hit;
			}

			// Step to time of reaching boundary to next cell, set time for that axis
			// to the next boundary intersection, and step current cell.
			debug_assert!(next.x >= 0.0);
			debug_assert!(next.y >= 0.0);
			debug_assert!(next.z >= 0.0);
			if direction.x != 0.0 && next.x <= next.y && next.x <= next.z {
				time += next.x;
				next.y -= next.x;
				next.z -= next.x;
				next.x = step.x;
				cell[0] += sign[0];
			} else if direction.y != 0.0 && next.y <= next.x && next.y <= next.z {
				time += next.y;
				next.x -= next.y;
				next.z -= next.y;
				next.y = step.y;
				cell[1] += sign[1];
			} else if direction.z != 0.0 && next.z <= next.x && next.z <= next.y {
				time += next.z;
				next.x -= next.z;
				next.y -= next.z;
				next.z = step.z;
				cell[2] += sign[2];
			} else {
				unreachable!();
			}
		}
	}

	fn time_to_aabb(start: F3, direction: F3, origin: F3, size: F3) -> f32 {
		// Time until reaching outside face for each axis.
		let time = f3(
			Self::time_to_aabb_boundary(start.x, direction.x, origin.x, size.x),
			Self::time_to_aabb_boundary(start.y, direction.y, origin.y, size.y),
			Self::time_to_aabb_boundary(start.z, direction.z, origin.z, size.z),
		);
		// Choose nearest face.
		if time.x >= 0.0 && (time.x <= time.y || time.y < 0.0) && (time.x <= time.z || time.z < 0.0) {
			time.x
		} else if time.y >= 0.0 && (time.y <= time.x || time.x < 0.0) && (time.y <= time.z || time.z < 0.0) {
			time.y
		} else if time.z >= 0.0 && (time.z <= time.x || time.x < 0.0) && (time.z <= time.y || time.y < 0.0) {
			time.z
		} else {
			0.0
		}
	}
	fn time_to_aabb_boundary(start: f32, direction: f32, origin: f32, size: f32) -> f32 {
		if direction > 0.0 {
			(origin-start)/direction
		} else if direction < 0.0 {
			-(start-origin-size)/direction
		} else {
			-1.0
		}
	}
	fn time_to_boundary(start: f32, direction: f32, grid_origin: f32, cell_size: f32) -> f32 {
		let scaled = (start-grid_origin)/cell_size;
		let fraction = (scaled-scaled.floor())*cell_size;
		if direction > 0.0 {
			(cell_size-fraction)/direction
		} else if direction < 0.0 {
			-fraction/direction
		} else {
			f32::INFINITY
		}
	}
	fn debug_aabb(debug: &mut DynamicMesh<engine::Vertex>, origin: F3, size: F3) {
		// Bottom edges.
		let vertex0 = debug_vertex(debug, origin);
		let vertex1 = debug_vertex(debug, origin+f3(size.x, 0.0, 0.0));
		debug.repeat(vertex1); let vertex2 = debug_vertex(debug, origin+f3(size.x, 0.0, size.z));
		debug.repeat(vertex2); let vertex3 = debug_vertex(debug, origin+f3(0.0, 0.0, size.z));
		debug.repeat(vertex3); debug.repeat(vertex0);
		// Top edges;
		let vertex4 = debug_vertex(debug, origin+f3(0.0, size.y, 0.0));
		let vertex5 = debug_vertex(debug, origin+f3(size.x, size.y, 0.0));
		debug.repeat(vertex5); let vertex6 = debug_vertex(debug, origin+f3(size.x, size.y, size.z));
		debug.repeat(vertex6); let vertex7 = debug_vertex(debug, origin+f3(0.0, size.y, size.z));
		debug.repeat(vertex7); debug.repeat(vertex4);
		// Middle edges.
		debug.repeat(vertex0); debug.repeat(vertex4);
		debug.repeat(vertex1); debug.repeat(vertex5);
		debug.repeat(vertex2); debug.repeat(vertex6);
		debug.repeat(vertex3); debug.repeat(vertex7);
	}
	fn debug_line(debug: &mut DynamicMesh<engine::Vertex>, start: F3, end: F3) {
		debug_vertex(debug, start); debug_vertex(debug, end);
	}
	fn debug_marker(debug: &mut DynamicMesh<engine::Vertex>, centre: F3) {
		let radius = 1.0;
		debug_vertex(debug, centre+f3(-radius, -radius, -radius));
		debug_vertex(debug, centre+f3(radius, radius, radius));
		debug_vertex(debug, centre+f3(radius, -radius, -radius));
		debug_vertex(debug, centre+f3(-radius, radius, radius));
		debug_vertex(debug, centre+f3(-radius, -radius, radius));
		debug_vertex(debug, centre+f3(radius, radius, -radius));
		debug_vertex(debug, centre+f3(radius, -radius, radius));
		debug_vertex(debug, centre+f3(-radius, radius, -radius));
	}
	fn debug_triangle(debug: &mut DynamicMesh<engine::Vertex>, vertices: [F3; 3]) {
		let vertex0 = debug_vertex(debug, vertices[0]);
		let vertex1 = debug_vertex(debug, vertices[1]);
		debug.repeat(vertex1); let vertex2 = debug_vertex(debug, vertices[2]);
		debug.repeat(vertex2); debug.repeat(vertex0);
	}
}

fn debug_vertex(debug: &mut DynamicMesh<engine::Vertex>, position: F3) -> Option<Index> {
	debug.push(engine::Vertex {
		position: position.to_array(),
		normal: [0.0; 3],
		texcoord: [0.0; 2],
	})
}

/// Information about a ray-mesh intersection.
pub struct RayHit {
	/// Ray direction, not necessarily a unit vector.
	pub direction: F3,
	/// Normal of face hit (unit vector).
	pub normal: F3,
	/// Start of ray.
	pub start: F3,
	/// Time to intersection along ray (also distance if `direction` has length of one).
	pub time: f32,
}
impl RayHit {
	/// Return position of intersection in local coordinates.
	pub fn point(&self) -> F3 { self.start+self.direction*self.time }
}

struct Cell {
	triangles: Box<[Triangle]>,
}

/// Pre-calculated collision data for a triangle.
/// Ray queries use Havel and Herout's algorithm.
/// See: Yet Faster Ray-Triangle Intersection (Using SSE4) (Havel and Herout, 2010)
/// See: https://stackoverflow.com/questions/13163129/ray-triangle-intersection
#[derive(Clone)]
struct Triangle {
	/// `normals[0].length().recip()` to let us return normals as unit vectors
	/// cheaply.
	normal_recip: f32,
	/// Normal for triangle and two edge-planes scaled for use in Havel and Herout.
	/// AKA N, N1, N2 in the paper.
	/// Note that the triangle's normal (element zero) points inward and is not a
	/// unit vector.
	normals: [F3; 3],
	/// Constants combined with `normals` to describe planes.
	/// Called d, d1, d2 in the paper.
	offsets: [f32; 3],
	/// Position of each corner used for debug graphics.
	#[cfg(debug_assertions)]
	vertices: [F3; 3],
}
impl Triangle {
	pub fn new(vertices: [F3; 3]) -> Self {
		let u = vertices[1]-vertices[0]; // AKA A->B
		let v = vertices[2]-vertices[0]; // AKA A->C
		let normal = u.cross(v);
		let length2 = normal.length2();
		assert!(length2 > 0.0, "degenerate (flat) triangle in collision mesh");
		let recip_length2 = length2.recip();
		let n1 = v.cross(normal)*recip_length2;
		let n2 = normal.cross(u)*recip_length2;
		Self {
			normal_recip: recip_length2.sqrt(),
			normals: [normal, n1, n2],
			offsets: [
				vertices[0].dot(normal),
				-vertices[0].dot(n1),
				-vertices[0].dot(n2),
			],
			#[cfg(debug_assertions)]
			vertices,
		}
	}
	pub fn ray(&self, start: F3, direction: F3, range: Option<f32>) -> Option<RayHit> {
		// This algorithm is built around these concepts:
		//
		// Planes can be described as a normal and an offset.
		// Point `p` is on the plane with normal `n` and offset `d` when:
		// p·n = d
		// 
		// Barycentric coordinates make it easy to tell positions inside a triangle.
		// They can be calculated by projecting (dot product) a cartesian point onto
		// directions perpendicular to the edge of the triangle and dividing by the
		// distance of the point opposite that face.
		//
		// Combining projecting the ray onto the triangle's normal and onto vectors
		// perpendicular to the edges of the triangle (but in its plane) allows
		// finding the point where the ray hits the triangle's plane, and from that
		// the barycentric coordinates of the hit.
		//
		// This is made more efficient by rearranging the calculations so that the
		// division needed in finding the time from its start that the ray hits the
		// plane happens as late as possible (it can often be skipped as a result).
		// All of the vectors used for projection (`self.normals`) are also scaled
		// so that the dot product does both the projection and division steps of
		// the conversion to barycentric coordinates.

		// Every `almost_<value>` is `<value>*determinant`, so we can check for
		// collision without dividing (which is expensive).
		let determinant = direction.dot(self.normals[0]);
		let limit = 0.000001; // complete guess
		if determinant.abs() < limit {
			// Ray parallel to triangle, likely to cause division by zero.
			return None;
		}
		let almost_time = self.offsets[0]-start.dot(self.normals[0]); // AKA t'
		if almost_time.is_sign_positive() != determinant.is_sign_positive() {
			// Hit before start of ray.
			return None;
		}
		debug_assert!(range.is_none() || range.unwrap() > 0.0);
		if range.is_some() && almost_time.abs() > range.unwrap()*determinant.abs() {
			// Hit after end of ray.
			return None;
		}
		let almost_point = start*determinant+direction*almost_time; // AKA P'
		let almost_u = almost_point.dot(self.normals[1])+determinant*self.offsets[1]; // AKA u'
		let almost_v = almost_point.dot(self.normals[2])+determinant*self.offsets[2]; // AKA v'
		if almost_u.is_sign_positive() != determinant.is_sign_positive() ||
			almost_v.is_sign_positive() != determinant.is_sign_positive() ||
			(almost_u+almost_v).abs() > determinant.abs()
		{
			// Hit outside of triangle.
			// `u` and `v` are barycentric coordinates, requiring the following for
			// inside:
			// - `u >= 0.0`
			// - `v >= 0.0`
			// - `u+v <= 1.0`
			// We can compensate for the multiplication by `determinant` by checking
			// signs for the first two cases, and multiplying the right hand side for
			// the second.
			return None;
		}
		// Remove the multiply by `determinant` now we know we hit the triangle.
		let determinant_recip = determinant.recip();
		let time = almost_time*determinant_recip;
		debug_assert!(time >= 0.0);
		debug_assert!(range.is_none() || time <= range.unwrap());
		let u = almost_u*determinant_recip;
		debug_assert!(u >= 0.0);
		let v = almost_v*determinant_recip;
		debug_assert!(v >= 0.0);
		debug_assert!(u+v <= 1.0);
		Some(RayHit {
			direction,
			normal: -self.normals[0]*self.normal_recip,
			start,
			time,
		})
	}
}
