//! This file is part of Flight.
//! 
//! Flight is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
//! 
//! Flight is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
//! You should have received a copy of the GNU General Public License along with Flight. If not, see <https://www.gnu.org/licenses/>.

use ffi_wren::{WrenHandle, WrenVM, wrenReleaseHandle};
use engine::{
	ActiveShader, Anchor, Blend, Camera, CameraUniforms, F3, Rotation, StaticMesh,
	Uniform, Vertex, Window,
};
use std::cell::RefCell;
use std::mem::MaybeUninit;
use std::rc::{Rc, Weak};

pub type Handle = Weak<RefCell<Entity>>;

pub struct Entities {
	buckets: [Bucket; Self::DELAY as usize],
	buffer: Vec<Handle>,
	counter: usize,
}
impl Entities {
	const FPS: u32 = 10;
	const DELAY: u32 = engine::Window::FPS as u32/Self::FPS;
	const NEAR: f32 = 50.0;
	const FAR: f32 = 100.0;

	pub fn new() -> Self {
		let mut buckets: [MaybeUninit<Bucket>; Self::DELAY as usize] = unsafe {
			MaybeUninit::uninit().assume_init()
		};
		for bucket in buckets.iter_mut() {
			*bucket = MaybeUninit::new(Bucket {
				active: Vec::new(),
				inactive: Vec::new(),
			});
		}
		Self {
			buckets: unsafe{std::mem::transmute::<
				[MaybeUninit<Bucket>; Self::DELAY as usize],
				[Bucket; Self::DELAY as usize]>(buckets)},
			buffer: Vec::new(),
			counter: 0,
		}
	}
	pub fn tick<F: Fn(F3) -> f32>(&mut self, tick_queue: &mut Vec<*mut WrenHandle>, player_distance_squared: F) {
		// Update active state of entities in current bucket.
		// Discard dead handles.
		debug_assert!(self.buffer.len() == 0);
		let far2 = Self::FAR*Self::FAR;
		let bucket = &mut self.buckets[self.counter];
		let mut index = 0;
		while index < bucket.active.len() {
			if let Some(entity) = bucket.active[index].upgrade() {
				let entity = entity.borrow();
				let distance2 = player_distance_squared(entity.next_translate);
				if distance2 <= far2 {
					index += 1;
				} else {
					let handle = bucket.active.swap_remove(index);
					// Use a temporary list to stored moved handles.
					// If we move them to the inactive list immediately we'll process them a
					// asecond time.
					self.buffer.push(handle);
				}
			} else {
				bucket.active.swap_remove(index);
			}
		}
		let mut index = 0;
		while index < bucket.inactive.len() {
			if let Some(entity) = bucket.inactive[index].upgrade() {
				let entity = entity.borrow();
				let distance2 = player_distance_squared(entity.next_translate);
				if distance2 > far2 {
					index += 1;
				} else {
					let handle = bucket.inactive.swap_remove(index);
					bucket.active.push(handle);
				}
			} else {
				bucket.inactive.swap_remove(index);
			}
		}
		bucket.inactive.append(&mut self.buffer);
		// Update active entities in this bucket.
		debug_assert!(tick_queue.len() == 0);
		for entity in bucket.active.iter() {
			if let Some(entity) = entity.upgrade() { // entity logic may remove entities
				let mut entity = entity.borrow_mut();
				entity.anchor = Anchor::new(entity.scale, entity.next_rotate, entity.next_translate);
				entity.last_rotate = entity.next_rotate;
				entity.last_translate = entity.next_translate;
				if !entity.tick.is_null() {
					tick_queue.push(entity.tick);
				}
			}
		}
		// Interpolate properties for active entities in remaining buckets.
		for bucket in 0..self.buckets.len() {
			let age = if bucket < self.counter {
				self.counter-bucket
			} else {
				self.counter+self.buckets.len()-bucket
			};
			let factor = age as f32/self.buckets.len() as f32;
			if bucket != self.counter {
				for entity in self.buckets[bucket].active.iter() {
					if let Some(entity) = entity.upgrade() {
						let mut entity = entity.borrow_mut();
						let rotate = entity.last_rotate.fast_lerp(entity.next_rotate, factor);
						let translate = entity.last_translate.lerp(entity.next_translate, factor);
						entity.anchor = Anchor::new(entity.scale, rotate, translate);
					}
				}
			}
		}
		// Switch bucket.
		self.counter += 1;
		if self.counter >= self.buckets.len() {
			self.counter = 0;
		}
	}
	pub fn draw(&self, window: &mut Window, shader: &ActiveShader<Vertex>, shader_alpha: &Option<Uniform>, shader_camera: &CameraUniforms, camera: &Camera, meshes: &[&StaticMesh<Vertex>]) {
		debug_assert!(Self::FAR > Self::NEAR);
		let near2 = Self::NEAR*Self::NEAR;
		let far2 = Self::FAR*Self::FAR;
		let alpha_scale = (far2-near2).recip();
		Blend::Alpha.draw(window);
		for bucket in self.buckets.iter() {
			for entity in bucket.active.iter() {
				if let Some(entity) = entity.upgrade() {
					let entity = entity.borrow();
					let distance2 = (entity.anchor.translate-camera.translate()).length2();
					let alpha = (1.0-(distance2-near2)*alpha_scale).max(0.0).min(1.0);
					shader.set(window, shader_alpha, alpha);
					shader_camera.set_anchor(window, shader, camera, &entity.anchor, false);
					meshes[entity.mesh].draw(window, shader);
				}
			}
		}
	}
	pub fn push(&mut self, mesh: usize, scale: F3, rotate: Rotation, translate: F3) -> Handle {
		let entity = Rc::new(RefCell::new(Entity {
			anchor: Anchor::new(scale, rotate, translate),
			last_rotate: rotate,
			last_translate: translate,
			mesh,
			next_rotate: rotate,
			next_translate: translate,
			scale,
			tick: std::ptr::null_mut(),
			this: None,
		}));
		entity.borrow_mut().this = Some(Rc::clone(&entity));
		self.buckets[self.counter].inactive.push(Rc::downgrade(&entity));
		Rc::downgrade(&entity)
	}
	pub unsafe fn bind(&mut self, vm: *mut WrenVM, entity: &Handle, tick: *mut WrenHandle) {
		if let Some(entity) = entity.upgrade() {
			let mut entity = entity.borrow_mut();
			if !entity.tick.is_null() {
				wrenReleaseHandle(vm, entity.tick);
			}
			entity.tick = tick;
		} else if !tick.is_null() {
			wrenReleaseHandle(vm, tick);
		}
	}
	pub fn unbind_vm(&mut self, vm: *mut WrenVM) {
		for bucket in self.buckets.iter() {
			for entity in bucket.active.iter().chain(bucket.inactive.iter()) {
				if let Some(entity) = entity.upgrade() {
					let mut entity = entity.borrow_mut();
					if !entity.tick.is_null() { unsafe {
						wrenReleaseHandle(vm, entity.tick);
						entity.tick = std::ptr::null_mut();
					} }
				}
			}
		}
	}
	pub fn remove(&mut self, vm: *mut WrenVM, entity: &Handle) {
		if let Some(entity) = entity.upgrade() {
			let mut entity = entity.borrow_mut();
			if !entity.tick.is_null() { unsafe {
				wrenReleaseHandle(vm, entity.tick);
				entity.tick = std::ptr::null_mut();
			} }
			entity.this = None;
		}
	}
}

#[derive(Clone)]
pub struct Entity {
	anchor: Anchor,
	last_rotate: Rotation,
	last_translate: F3,
	pub mesh: usize,
	pub next_rotate: Rotation,
	pub next_translate: F3,
	pub scale: F3,
	tick: *mut WrenHandle,
	this: Option<Rc<RefCell<Self>>>,
}

#[derive(Clone)]
struct Bucket {
	active: Vec<Handle>,
	inactive: Vec<Handle>,
}
