//! This file is part of Flight.
//! 
//! Flight is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
//! 
//! Flight is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
//! You should have received a copy of the GNU General Public License along with Flight. If not, see <https://www.gnu.org/licenses/>.

extern crate engine;

mod aircraft;
mod bitmap;
mod camera;
mod collision;
mod entities;
mod ffi_wren;
mod game;
mod heightmap;
mod input;
mod scripting;
mod shadow;
mod trail;

use aircraft::{Aircraft, Surface};
use bitmap::Bitmap;
use camera::{FollowCamera, FPSCamera};
use collision::CollisionMesh;
use engine::*;
use entities::Entities;
use game::{GameData, SeatData};
use heightmap::{MapCell, ONeilHeightmap};
use input::{Input, LogicalAxis, Seat};
use scripting::Scripting;
use shadow::{Fill, Shadow};
use trail::Trail;
use std::cell::RefCell;
use std::time::Instant;

// Benchmark with no other activity for half display.
// Before SceneryMesh:
// CPU ~60%
// rcs0 ~100%
// 387 FPS
// 99.3k geometry
// 77.9k textures
// After SceneryMesh:
// CPU ~60%
// rcs0 ~100%
// 366 FPS
// 1,771.1k geometry
// 77.9k textures

fn main() {
	let mut window = Window::new("Flight").unwrap();
	let mut tweak = Tweak::new();
	let mut game = GameData {
		entities: Entities::new(),
		floor_collision: CollisionMesh::invalid(),
		input: Input::new(4),
		rng: RNG::new(),
	};
	let mut scripting = Scripting::new();
	if let Ok(script) = std::fs::read("/tmp/test.wren") {
		let script = String::from_utf8(script).unwrap();
		scripting.run(&script, &mut game);
	}

	let shaders = Shaders::load(&mut window);
	let textures = Textures::load(&mut window);

	let mut debug_camera = FPSCamera::new(0.1, 200.0, 60.0/180.0*std::f32::consts::PI, window.size(), f3(0.0, 0.0, -1.0));
	let mut controls = f3(0.0, 0.0, 0.0);
	let mut gravity = 1.0 as f32;
	let mut power_pitch = 0.0004 as f32;
	let mut power_roll = 0.0008 as f32;
	let mut power_yaw = 0.0001 as f32;
	let mut thrust_min = 0.0 as f32;
	let mut thrust_half = 1.0 as f32;
	let mut thrust_max = 1.5 as f32;
	let mut surfaces = [
		Surface{area: 0.1, centre: f3(0.0, 0.0, -0.00015), normal: f3(1.0, 0.0, 0.0)},
		Surface{area: 0.4, centre: f3(0.0, 0.0, 0.0), normal: f3(0.0, 1.0, 0.0)},
		Surface{area: 0.05, centre: f3(0.0, 0.0, 0.0), normal: f3(0.0, 0.0, 1.0)},
	];
	let mut trail_min = 0.2;
	let mut trail_amount = 0.0003;
	let mut spin = 0.0;

	let heightmap_radius = 128.0;
	let aircraft_data = MeshData::embedded(include_bytes!("../assets/aircraft.mesh"));
	let mut meshes = {
		// let heightmap = Bitmap::embedded(include_bytes!("../assets/mountain.rgba"));
		let heightmap = Bitmap::read(std::fs::File::open("assets/mountain.rgba").unwrap()).unwrap();
		let size = heightmap.width() as f32;
		let heightmap = ONeilHeightmap::new(&heightmap, heightmap_radius, 0.5);

		let mut floor_data = heightmap.to_mesh();
		// Perturb texture coordinates of landscape to make patterns less regular.
		// This is mostly useful for land.
		for vertex in floor_data.vertices.iter_mut() {
			vertex.texcoord.iter_mut().for_each(|value| *value += (game.rng.f32()*2.0-1.0)*0.5/size);
		}
		let mut builder = SceneryBuilder::new();
		builder.push(&floor_data.indices, &floor_data.vertices,
			f3(1.0, 1.0, 1.0), Rotation::identity(), f3(0.0, 0.0, 0.0));
		let floor = builder.build(&mut window, [16, 2, 2], 180.0, 20.0);

		let mut clouds_data = heightmap.to_sea_mesh(0.5, 4.0);
		// Turn inside out.
		clouds_data.indices.chunks_exact_mut(3).for_each(|triangle| triangle.swap(1, 2));

		let mut collision_buffer = SceneryBuilder::new();
		collision_buffer.push(&floor_data.indices, &floor_data.vertices,
			f3(1.0, 1.0, 1.0), Rotation::identity(), f3(0.0, 0.0, 0.0));

		let archway_data = MeshData::embedded(include_bytes!("../assets/archway.mesh"));
		let archway = StaticMesh::new(&mut window,
			Primitive::Triangles, &archway_data.indices, &archway_data.vertices);
		let arrow_data = MeshData::embedded(include_bytes!("../assets/arrow.mesh"));
		let arrow = StaticMesh::new(&mut window,
			Primitive::Triangles, &arrow_data.indices, &arrow_data.vertices);
		let finger_data = MeshData::embedded(include_bytes!("../assets/finger.mesh"));
		let finger = StaticMesh::new(&mut window,
			Primitive::Triangles, &finger_data.indices, &finger_data.vertices);
		let gull_data = MeshData::embedded(include_bytes!("../assets/gull.mesh"));
		let gull = StaticMesh::new(&mut window,
			Primitive::Triangles, &gull_data.indices, &gull_data.vertices);
		let ring_data = MeshData::embedded(include_bytes!("../assets/ring.mesh"));
		let ring = StaticMesh::new(&mut window,
			Primitive::Triangles, &ring_data.indices, &ring_data.vertices);
		let shrub_data = MeshData::embedded(include_bytes!("../assets/shrub.mesh"));
		let shrub = StaticMesh::new(&mut window,
			Primitive::Triangles, &shrub_data.indices, &shrub_data.vertices);
		let tree_data = MeshData::embedded(include_bytes!("../assets/tree.mesh"));
		let tree = StaticMesh::new(&mut window,
			Primitive::Triangles, &tree_data.indices, &tree_data.vertices);

		let items = &[
			&archway_data,
			&finger_data,
			&shrub_data,
			&tree_data,
		];
		let mut builder = SceneryBuilder::new();
		heightmap.for_triangles(|cell| for _ in 0..8 {
			let mut u = game.rng.f32();
			let mut v = game.rng.f32();
			if u+v > 1.0 {
				u = 1.0-u;
				v = 1.0-v;
			}
			let height = cell.height(u, v);
			if height > 0.2 && height < 0.5 {
				let angle = MapCell::interpolate_f32([
					cell.points[0].uv.y,
					cell.points[1].uv.y,
					cell.points[2].uv.y,
				], u, v)*std::f32::consts::PI*2.0;
				let data = game.rng.choose(items);
				let scale = f3(2.0, 2.0, 2.0);
				let rotation = Rotation::new(f3(0.0, 1.0, 0.0), game.rng.f32()*std::f32::consts::PI*2.0)
					.then(Rotation::new(f3(1.0, 0.0, 0.0), angle));
				let translate = cell.position(u, v);
				builder.push(&data.indices, &data.vertices, scale, rotation, translate);
				collision_buffer.push(&data.indices, &data.vertices, scale, rotation, translate);
			}
		});
		let scenery = builder.build(&mut window, [16, 2, 2], 100.0, 20.0);

		game.floor_collision = CollisionMesh::new(
			&collision_buffer.indices, &collision_buffer.vertices, [32, 16, 16]);

		Meshes {
			aircraft: aircraft_data.to_mesh(&mut window),
			archway,
			arrow,
			clouds: clouds_data.to_mesh(&mut window),
			fill: Fill::new(&mut window),
			finger,
			floor,
			gull,
			ring,
			scenery,
			shadow: Shadow::new(&mut window, &aircraft_data),
			shrub,
			tree,
		}
	};

	let mut debug = DynamicMesh::new(&mut window, Primitive::Lines);

	let benchmark = false;
	let mut benchmark_frames = 0;
	let benchmark_start = Instant::now();
	if benchmark {
		window.benchmark().unwrap();
	}

	let mut tick_queue = Vec::new();
	'main_loop: loop {
		let mut viewports_changed = false;
		let mut viewports_changed_alternate = false;
		while let Some(event) = window.poll() {
			game.input.event(&event, || {
				viewports_changed = true;
				SeatData {
					aircraft: Aircraft {
						centre: f3(0.0, 0.0, 0.0),
						crashed_counter: 0,
						mass_recip: 1.0,
						rotation: Rotation::identity(),
						spin: f3(0.0, 0.0, 0.0),
						spin_damping: 0.99,
						spin_recip: 1.0,
						velocity: f3(0.0, 0.0, 1.0),
					},
					camera: FollowCamera::new(0.1, 200.0, 60.0/180.0*std::f32::consts::PI, window.size(), f3(0.0, 0.0, -1.0)),
					shadow: RefCell::new(meshes.shadow.clone(&mut window)),
					trail_alpha: 0.0,
					trail_left: RefCell::new(Trail::new(&mut window, f3(0.0, 0.0, 0.0))),
					trail_right: RefCell::new(Trail::new(&mut window, f3(0.0, 0.0, 0.0))),
				}
			}, |_| viewports_changed_alternate = true);
			match event {
				Event::Key{key, down, modifiers, ..} => {
					let value = if down { 1.0 } else { 0.0 };
					if key == Key::Q && modifiers.alt {
						// Emergency quit, mostly for Raspberry Pi console use.
						break 'main_loop;
					} else if key == Key::W {
						controls.x = value;
					} else if key == Key::S {
						controls.x = -value;
					} else if key == Key::A {
						controls.z = value;
					} else if key == Key::D {
						controls.z = -value;
					} else if key == Key::Q {
						controls.y = -value;
					} else if key == Key::E {
						controls.y = value;
					}
				},
				Event::Motion{position: _, relative} => if game.input.seats().len() == 0 {
					debug_camera.rotate_y += relative.x*0.01;
					debug_camera.rotate_x -= relative.y*0.01;
				},
				Event::Quit => break 'main_loop,
				Event::Resize => {
					viewports_changed = true;
					debug_camera.camera.viewport(f2(0.0, 0.0), window.size());
				},
				_ => (),
			}
		}
		if (viewports_changed || viewports_changed_alternate) && game.input.seats().len() > 0 {
			const LAYOUTS: [&'static [(F2, F2)]; 4] = [
				&[(f2(0.0, 0.0), f2(1.0, 1.0))],
				&[(f2(0.0, 0.0), f2(0.5, 1.0)), (f2(0.5, 0.0), f2(0.5, 1.0))],
				&[(f2(0.0, 0.5), f2(0.5, 0.5)), (f2(0.5, 0.5), f2(0.5, 0.5)),
					(f2(0.0, 0.0), f2(0.5, 0.5)), (f2(0.5, 0.0), f2(0.5, 0.5))],
				&[(f2(0.0, 0.5), f2(0.5, 0.5)), (f2(0.5, 0.5), f2(0.5, 0.5)),
					(f2(0.0, 0.0), f2(0.5, 0.5)), (f2(0.5, 0.0), f2(0.5, 0.5))],
			];
			let layout = LAYOUTS[game.input.seats().len()-1];
			let window = window.size();
			for (seat, (topleft, size)) in game.input.seats_mut().iter_mut().zip(layout.iter().copied()) {
				seat.data.camera.camera.viewport(topleft*window, size*window);
			}
		}
		tweak.f32("gravity", &mut gravity);
		tweak.f32("pitch", &mut power_pitch);
		tweak.f32("roll", &mut power_roll);
		tweak.f32("yaw", &mut power_yaw);
		tweak.f32("thrust_min", &mut thrust_min);
		tweak.f32("thrust_half", &mut thrust_half);
		tweak.f32("thrust_max", &mut thrust_max);
		tweak.f32("x.area", &mut surfaces[0].area);
		tweak.f3("x.centre", &mut surfaces[0].centre);
		tweak.f3("x.normal", &mut surfaces[0].normal);
		tweak.f32("y.area", &mut surfaces[1].area);
		tweak.f3("y.centre", &mut surfaces[1].centre);
		tweak.f3("y.normal", &mut surfaces[1].normal);
		tweak.f32("z.area", &mut surfaces[2].area);
		tweak.f3("z.centre", &mut surfaces[2].centre);
		tweak.f3("z.normal", &mut surfaces[2].normal);
		tweak.f32("trail.min", &mut trail_min);
		tweak.f32("trail.amount", &mut trail_amount);
		let mut sea_fade = 30.0;
		tweak.f32("sea.fade", &mut sea_fade);
		tweak.tick();
		// HACK spin += 0.0003;
		// TODO: figure out answer to clouds so this value isn't being shared.
		spin += 3.14159/60.0*0.1;
		if spin > std::f32::consts::PI*2.0 { spin -= std::f32::consts::PI*2.0; }
		if game.input.seats().len() == 0 {
			debug_camera.tick();
			let velocity = debug_camera.camera.rotate().apply(f3(-controls.z, 0.0, 0.0))
				+debug_camera.camera.rotate().apply(f3(0.0, controls.y, 0.0))
				+debug_camera.camera.rotate().apply(f3(0.0, 0.0, controls.x));
			debug_camera.centre += velocity;
		}
		for seat in game.input.seats_mut().iter_mut() {
			seat.data.camera.tick(seat.data.aircraft.centre, seat.data.aircraft.rotation);
			seat.data.aircraft.spin += seat.data.aircraft.rotation.apply(
				f3(seat.axis(LogicalAxis::Pitch)*power_pitch, 0.0, 0.0));
			seat.data.aircraft.spin += seat.data.aircraft.rotation.apply(
				f3(0.0, seat.axis(LogicalAxis::Yaw)*power_yaw, 0.0));
			seat.data.aircraft.spin += seat.data.aircraft.rotation.apply(
				f3(0.0, 0.0, -seat.axis(LogicalAxis::Roll)*power_roll));
			let throttle = seat.axis(LogicalAxis::Throttle);
			let thrust = if throttle < 0.0 { (thrust_half-thrust_min)*throttle } else { (thrust_max-thrust_half)*throttle }+thrust_half;
			seat.data.aircraft.velocity += seat.data.aircraft.rotation.apply(f3(0.0, 0.0, thrust));
			let centre = seat.data.aircraft.centre;
			let direction = f3(0.0, centre.y, centre.z).unit_zero();
			// seat.data.aircraft.velocity += f3(0.0, seat.data.aircraft.centre.y, seat.data.aircraft.centre.z)*gravity/heightmap_radius;
			seat.data.aircraft.velocity += direction*gravity;
			seat.data.aircraft.tick(&surfaces, &game.floor_collision);
			seat.data.shadow.borrow_mut().update(f3(1.0, 1.0, 1.0), seat.data.aircraft.rotation, seat.data.aircraft.centre, f3(seat.data.aircraft.centre.x, 0.0, 0.0), 100.0);
			let left = seat.data.aircraft.rotation.apply(f3(-1.0, 0.0, 0.039));
			seat.data.trail_alpha = (seat.data.aircraft.velocity.length2()*trail_amount-trail_min)
				.max(0.0)
				.min(1.0);
			seat.data.trail_left.borrow_mut().update(left+seat.data.aircraft.centre);
			let right = seat.data.aircraft.rotation.apply(f3(1.0, 0.0, 0.039));
			seat.data.trail_right.borrow_mut().update(right+seat.data.aircraft.centre);
		}
		{
			let entities = &mut game.entities;
			let input = &game.input;
			entities.tick(&mut tick_queue, |point: F3| {
				let mut distance2 = (point-debug_camera.centre).length2();
				for seat in input.seats().iter() {
					distance2 = distance2.min((seat.data.aircraft.centre-point).length2());
				}
				distance2
			});
		}
		for closure in tick_queue.drain(..) { unsafe {
			scripting.call_closure(&mut game, closure);
		} }
		scripting.tick(&mut game);
		rgba(0xebebf7ff).draw(&mut window);
		if game.input.seats().len() == 0 {
			draw_viewport(&mut window, &debug_camera.camera, &shaders, &textures, &mut meshes, game.input.seats(), &game.entities, spin, &mut debug);
		} else {
			for seat in game.input.seats().iter() {
				draw_viewport(&mut window, &seat.data.camera.camera, &shaders, &textures, &mut meshes, game.input.seats(), &game.entities, spin, &mut debug);
			}
		}
		window.frame();
		debug.clear();

		if benchmark {
			benchmark_frames += 1;
			let frames = Window::FPS as u32*30;
			if benchmark_frames >= frames {
				let seconds = benchmark_start.elapsed().as_secs_f32();
				let fps = frames as f32/seconds;
				println!("Benchmark: {} frames in {} seconds", frames, seconds);
				println!("Benchmark: {} FPS ({}x)", fps, fps/Window::FPS);
				break 'main_loop;
			}
		}
	}

	unsafe {
		game.entities.unbind_vm(scripting.vm());
	}
}

fn draw_viewport(window: &mut Window, camera: &Camera, shaders: &Shaders, textures: &Textures, meshes: &mut Meshes, seats: &[Seat<SeatData>], entities: &Entities, cloud_spin: f32, debug: &mut DynamicMesh<Vertex>) {
	window.depth(true);
	camera.draw(window);
	// Land and sea.
	let shader = shaders.floor.draw(window);
	shaders.floor_camera.set_identity(window, &shader, camera, false);
	shader.set(window, &shaders.floor_heightmap, (0, &textures.floor_heightmap));
	shader.set(window, &shaders.floor_pattern, (2, &textures.floor_pattern));
	shader.set(window, &shaders.floor_sea_noise, (1, &textures.floor_sea_noise));
	shader.set(window, &shaders.floor_sea_sin, cloud_spin.sin());
	shader.set(window, &shaders.floor_sea_cos, cloud_spin.cos());
	debug_assert!(textures.floor_heightmap.width() == textures.floor_heightmap.height());
	shader.set(window, &shaders.floor_heightmap_size, textures.floor_heightmap.width());
	meshes.floor.draw(window, camera.translate(), &shader, &shaders.floor_alpha);
	Blend::Replace.draw(window);
	// Clouds.
	let shader = shaders.palette.draw(window);
	shaders.palette_camera.set(window, &shader, camera,
		f3(1.0, 1.0, 1.0), Rotation::new(f3(1.0, 0.0, 0.0), -cloud_spin), f3(0.0, 0.0, 0.0), false);
	shader.set(window, &shaders.palette_texture, (0, &textures.clouds));
	meshes.clouds.draw(window, &shader);
	// Aircraft.
	let palette_count = 4.0;
	shader.set(window, &shaders.palette_palette, (1.0+0.5)/palette_count);
	shader.set(window, &shaders.palette_texture, (0, &textures.palette));
	for seat in seats.iter() {
		let aircraft = &seat.data.aircraft;
		shaders.palette_camera.set(window, &shader, camera,
			f3(1.0, 1.0, 1.0), aircraft.rotation, aircraft.centre, false);
		meshes.aircraft.draw(window, &shader);
	}
	// Scenery.
	shader.set(window, &shaders.palette_palette, (3.0+0.5)/palette_count);
	shaders.palette_camera.set_identity(window, &shader, camera, false);
	meshes.scenery.draw(window, camera.translate(), &shader, &shaders.palette_alpha);
	// Entities.
	entities.draw(window, &shader, &shaders.palette_alpha, &shaders.palette_camera, camera, &[
		&meshes.aircraft,
		&meshes.archway,
		&meshes.arrow,
		&meshes.finger,
		&meshes.gull,
		&meshes.ring,
		&meshes.shrub,
		&meshes.tree,
	]);
	/*
	HACK
	// Motion specks.
	shaded_camera.set(window, &shader, camera,
		f3(1.0, 1.0, 1.0), Rotation::identity(), body.centre, true);
	lines.draw(window, &shader);
	*/
	// Shadows.
	let context = Shadow::draw_start(window);
	let shader = shaders.shadow_volume.draw(window);
	for seat in seats.iter() {
		shaders.shadow_volume_camera.set(window, &shader, camera,
			f3(1.0, 1.0, 1.0), seat.data.aircraft.rotation, seat.data.aircraft.centre, false);
		seat.data.shadow.borrow_mut().draw_add(window, &context, &shader);
	}
	Blend::Multiply.draw(window);
	shader.set(window, &shaders.shadow_volume_camera.clip_matrix, &Matrix3::identity());
	shader.set(window, &shaders.shadow_volume_camera.clip_translate, f3(0.0, 0.0, 0.0));
	shader.set(window, &shaders.shadow_volume_camera.clip_z_offset, 0.0);
	shader.set(window, &shaders.shadow_volume_camera.clip_z_scale, 1.0);
	Shadow::draw_end(window, context, &meshes.fill, &shader);
	// Trails.
	Blend::Alpha.draw(window);
	let shader = shaders.trail.draw(window);
	shaders.trail_camera.set_identity(window, &shader, camera, false);
	shader.set(window, &shaders.trail_camera_position, camera.translate());
	shader.set(window, &shaders.trail_rgba, [1.0, 0.0, 1.0, 1.0]);
	shader.set(window, &shaders.trail_width, 0.1);
	for seat in seats.iter() {
		if seat.data.trail_alpha > 0.0 {
			shader.set(window, &shaders.trail_rgba, [1.0, 1.0, 1.0, seat.data.trail_alpha]);
			seat.data.trail_left.borrow_mut().draw(window, &shader);
			seat.data.trail_right.borrow_mut().draw(window, &shader);
		}
	}
	// Debugging graphics.
	Blend::Replace.draw(window);
	let shader = shaders.debug.draw(window);
	shaders.debug_camera.set_identity(window, &shader, camera, false);
	debug.draw(window, &shader);
}

pub struct Shaders {
	debug: Shader<Vertex>,
	debug_camera: CameraUniforms,

	floor: Shader<Vertex>,
	floor_camera: CameraUniforms,
	floor_heightmap: Option<Uniform>,
	floor_pattern: Option<Uniform>,
	floor_heightmap_size: Option<Uniform>,
	floor_sea_noise: Option<Uniform>,
	floor_sea_sin: Option<Uniform>,
	floor_sea_cos: Option<Uniform>,
	floor_alpha: Option<Uniform>,

	palette: Shader<Vertex>,
	palette_alpha: Option<Uniform>,
	palette_camera: CameraUniforms,
	palette_palette: Option<Uniform>,
	palette_texture: Option<Uniform>,

	shadow_volume: Shader<Vertex>,
	shadow_volume_camera: CameraUniforms,

	trail: Shader<Vertex>,
	trail_camera: CameraUniforms,
	trail_camera_position: Option<Uniform>,
	trail_rgba: Option<Uniform>,
	trail_width: Option<Uniform>,
}
impl Shaders {
	pub fn load(window: &mut Window) -> Self {
		let debug = Shader::load(window, "debug", "shaders/debug").unwrap();
		let debug_camera = CameraUniforms::new(window, &debug, false);

		let floor = Shader::load(window, "floor", "shaders/floor").unwrap();
		let floor_camera = CameraUniforms::new(window, &floor, false);
		let floor_heightmap = floor.uniform(window, "texture_heightmap");
		let floor_pattern = floor.uniform(window, "texture_detail");
		let floor_sea_noise = floor.uniform(window, "texture_noise");
		let floor_heightmap_size = floor.uniform(window, "heightmap_size");
		let floor_sea_sin = floor.uniform(window, "sea_sin");
		let floor_sea_cos = floor.uniform(window, "sea_cos");
		let floor_alpha = floor.uniform(window, "alpha");

		let shadow_volume = Shader::load(window, "shadow", "shaders/shadow").unwrap();
		let shadow_volume_camera = CameraUniforms::new(window, &shadow_volume, false);

		let palette = Shader::load(window, "palette", "shaders/palette").unwrap();
		let palette_alpha = palette.uniform(window, "alpha");
		let palette_camera = CameraUniforms::new(window, &palette, false);
		let palette_palette = palette.uniform(window, "palette");
		let palette_texture = palette.uniform(window, "texture");

		let trail = Shader::load(window, "trail", "shaders/trail").unwrap();
		let trail_camera = CameraUniforms::new(window, &trail, false);
		let trail_camera_position = trail.uniform(window, "camera_position");
		let trail_rgba = trail.uniform(window, "trail_rgba");
		let trail_width = trail.uniform(window, "trail_width");

		Self {
			debug,
			debug_camera,

			floor,
			floor_camera,
			floor_heightmap,
			floor_pattern,
			floor_heightmap_size,
			floor_sea_noise,
			floor_sea_sin,
			floor_sea_cos,
			floor_alpha,

			palette,
			palette_alpha,
			palette_camera,
			palette_palette,
			palette_texture,

			shadow_volume,
			shadow_volume_camera,

			trail,
			trail_camera,
			trail_camera_position,
			trail_rgba,
			trail_width,
		}
	}
}

pub struct Textures {
	clouds: Texture,
	// Colours of landscape (land and sea) in RGB, heigh in alpha.
	// Colours alternate between variations in a checkerboard to make sampling of
	// adjacent texels return different colours.
	// This produces the mixture of colours used to draw detail onto land.
	//
	// The RGB must be present even in zero height areas.
	// GIMP discards transparent pixels, so export a separate copy of the image
	// with no alpha, and use Image Magick to comine the pieces:
	// `convert mountain_rgb.png mountain_alpha.png -compose CopyOpacity -composite mountain.bmp`
	floor_heightmap: Texture,
	// Pattern used for land and sea.
	// The land uses two layers of the pattern for large and small features.
	// The sea uses one layer and warps it to give a different feel.
	floor_pattern: Texture,
	// Noise used to distort sea shapes.
	floor_sea_noise: Texture,
	palette: Texture,
}
impl Textures {
	pub fn load(window: &mut Window) -> Self {
		Self {
			clouds: Texture::embedded(window, include_bytes!("../assets/clouds.rgba"), true, Repeat::Copy),
			floor_heightmap: Texture::embedded(window, include_bytes!("../assets/mountain.rgba"), false, Repeat::Copy),
			floor_pattern: Texture::embedded(window, include_bytes!("../assets/pattern.rgba"), false, Repeat::Mirror),
			floor_sea_noise: Texture::embedded(window, include_bytes!("../assets/noise.rgba"), true, Repeat::Copy),
			palette: Texture::embedded(window, include_bytes!("../assets/palette.rgba"), false, Repeat::None),
		}
	}
}

pub struct Meshes {
	aircraft: StaticMesh<Vertex>,
	archway: StaticMesh<Vertex>,
	arrow: StaticMesh<Vertex>,
	clouds: StaticMesh<Vertex>,
	fill: Fill,
	finger: StaticMesh<Vertex>,
	floor: SceneryMesh<Vertex>,
	gull: StaticMesh<Vertex>,
	ring: StaticMesh<Vertex>,
	scenery: SceneryMesh<Vertex>,
	shadow: Shadow,
	shrub: StaticMesh<Vertex>,
	tree: StaticMesh<Vertex>,
}

/*
pub struct Dust {
	camera: F3,
	data: Box<[F3]>,
	mesh: DynamicMesh<Vertex>,
}
impl Dust {
	const DISTANCE: f32 = 5.0;

	pub fn new(window: &mut Window, rng: &mut RNG) -> Self {
		let mut data = Vec::new()
		data.resize_with(64, || f3(
			rng.f32()*Self::DISTANCE*2.0-Self::DISTANCE,
			rng.f32()*Self::DISTANCE*2.0-Self::DISTANCE,
			rng.f32()*Self::DISTANCE*2.0-Self::DISTANCE,
		));
		Self {
			camera: f3(0.0, 0.0, 0.0),
			data: data.into_boxed_slice(),
			mesh: DynamicMesh::new(&mut window, Primitive::Lines),
		}
	}
	pub fn tick(&mut self, camera: F3) {
		let velocity = camera-self.camera;
		self.camera = camera;
		self.mesh.clear();
		for dot in self.data.iter_mut() {
			*dot -= velocity*0.01;
			if dot.x < -Self::DISTANCE { dot.x += Self::DISTANCE*2.0; }
			if dot.x > Self::DISTANCE { dot.x -= Self::DISTANCE*2.0; }
			if dot.y < -Self::DISTANCE { dot.y += Self::DISTANCE*2.0; }
			if dot.y > Self::DISTANCE { dot.y -= Self::DISTANCE*2.0; }
			if dot.z < -Self::DISTANCE { dot.z += Self::DISTANCE*2.0; }
			if dot.z > Self::DISTANCE { dot.z -= Self::DISTANCE*2.0; }
			self.mesh.push(Vertex {
				position: dot.to_array(),
				normal: [0.0; 3],
				texcoord: [0.0; 2],
			});
			self.mesh.push(Vertex {
				position: (*dot-velocity*0.01).to_array(),
				normal: [0.0; 3],
				texcoord: [0.0; 2],
			});
		}
	}
	pub fn draw(&mut self, window: &mut Window, shader: &ActiveShader<Vertex>) {
		self.mesh.draw_keep(window, shader);
	}
}
*/

/*
pub struct Scenery<'a, Vertex: VertexBinding> {
	fading_in: Vec<(SceneryItem<'a, Vertex>, f32)>,
	fading_out: Vec<(SceneryItem<'a, Vertex>, f32)>,
	invisible: VecDeque<SceneryItem<'a, Vertex>>,
	tick_index: usize,
	tick_items: usize,
	tick_visible: bool,
	visible: VecDeque<SceneryItem<'a, Vertex>>,
}
impl<'a, Vertex: VertexBinding> Scenery<'a, Vertex> {
	pub fn new() -> Self {
		Self {
			fading_in: Vec::new(),
			fading_out: Vec::new(),
			invisible: VecDeque::new(),
			tick_index: 0,
			tick_items: 1,
			tick_visible: true,
			visible: VecDeque::new(),
		}
	}
	pub fn push(&mut self, mesh: &'a StaticMesh<Vertex>, rotate: Rotation, centre: F3) {
		self.invisible.push_back(SceneryItem {
			anchor: Anchor::new(f3(0.25, 0.25, 0.25), rotate, centre),
			mesh,
		});
		let fps = 60;
		let update_seconds = 1;
		self.tick_items = ((self.fading_in.len()+self.fading_out.len()+self.invisible.len()+self.visible.len())*update_seconds/fps).max(1);
	}
	pub fn draw(&mut self, window: &mut Window, shader: &ActiveShader<Vertex>, camera_uniforms: &CameraUniforms, alpha_uniform: &Option<Uniform>, camera: &Camera) {
		// Show or hide things as they move into or out of view.
		let near2 = (100.0 as f32).powi(2);
		let far2 = (110.0 as f32).powi(2);
		for _ in 0..self.tick_items {
			if self.tick_visible {
				if self.tick_index >= self.visible.len() {
					self.tick_index = 0;
					self.tick_visible = false;
				} else {
					let distance2 = (self.visible[self.tick_index].anchor.translate-camera.translate()).length2();
					if distance2 > far2 {
						let item = self.visible.swap_remove_back(self.tick_index).unwrap();
						self.fading_out.push((item, 1.0));
					} else {
						self.tick_index += 1;
					}
				}
			} else {
				if self.tick_index >= self.invisible.len() {
					self.tick_index = 0;
					self.tick_visible = true;
				} else {
					let distance2 = (self.invisible[self.tick_index].anchor.translate-camera.translate()).length2();
					if distance2 < near2 {
						let item = self.invisible.swap_remove_back(self.tick_index).unwrap();
						self.fading_in.push((item, 0.0));
					} else {
						self.tick_index += 1;
					}
				}
			}
		}

		// Step fade animation.
		let step = 0.5/60.0;
		let mut index = 0;
		while index < self.fading_in.len() {
			self.fading_in[index].1 += step;
			if self.fading_in[index].1 >= 1.0 {
				let (item, _) = self.fading_in.swap_remove(index);
				self.visible.push_front(item);
				if self.tick_visible { self.tick_index += 1; }
			} else {
				index += 1;
			}
		}
		let mut index = 0;
		while index < self.fading_out.len() {
			self.fading_out[index].1 -= step;
			if self.fading_out[index].1 <= 0.0 {
				let (item, _) = self.fading_out.swap_remove(index);
				self.invisible.push_front(item);
				if !self.tick_visible { self.tick_index += 1; }
			} else {
				index += 1;
			}
		}

		// Draw visible decor.
		Blend::Replace.draw(window);
		shader.set(window, alpha_uniform, 1.0);
		for item in self.visible.iter() {
			camera_uniforms.set_anchor(window, shader, camera, &item.anchor, false);
			item.mesh.draw(window, shader);
		}
		Blend::Alpha.draw(window);
		for (item, alpha) in self.fading_in.iter().chain(self.fading_out.iter()) {
			camera_uniforms.set_anchor(window, shader, camera, &item.anchor, false);
			shader.set(window, alpha_uniform, *alpha);
			item.mesh.draw(window, shader);
		}
	}
}

pub struct SceneryItem<'a, Vertex: VertexBinding> {
	anchor: Anchor,
	mesh: &'a StaticMesh<Vertex>,
}
*/
